local myname="ZifnabsLocator"
local Me=CreateFrame("Frame",myname)
if not ZifnabsMod then ZifnabsMod={loc=Me} else ZifnabsMod.loc=Me end
local mod
local L
local C
local db

local FRACT=1000
local POS_MULTI=100*FRACT
local ZONE_MULTI=POS_MULTI*POS_MULTI

local DEFAULTSCALE={X=6.2,Y=4.8}
local ZOOMSCALE={[0]=1,[1]=1.1,[2]=1.3,[3]=1.7,[4]=2.3,[5]=3.3}
local MMMAXDIST=20
local MMMINDIST=0
local MMFADERANGE=0.1

local zone_name={}
local widget={mm={}}
local tick={tack=0,zone=nil,last=nil,sum=0,ix=1,list={}}
local unknownzone={}
local interested={zone={},map={}}

-- --------------------------------------------------------------------------------------
-- INITIALIZE
-- --------------------------------------------------------------------------------------
function Me:Initialize(fwinfo)
	mod=fwinfo.mod
	L=fwinfo.L
	C=fwinfo.C
	db=fwinfo.db
	if not db.locator then db.locator={mmscale={},speedometer=true,unknown={}} end
	
	for zone,nr in pairs(db.locator.unknown or {}) do
		if L.locator.zones[zone] then
			local changed=0
			for nodetype,data in pairs(db.gatherer.types) do
				if data.list then
					for listnr,entry in pairs(data.list) do
						local px,py,pzone,pzn=Me:ParseLoc(entry.pos)
						if pzn==nr then 
							local nx,ny,nz,npos,nnr=Me:GetLoc(pzone,px,py)
							entry.pos=npos
							changed=changed+1
						end
					end
				end
			end
			for name,data in pairs(db.targets.list) do
				for pos,whowhen in pairs(data.seen or {}) do
					local px,py,pzone,pzn=Me:ParseLoc(pos)
					if pzn==nr then
						local newpos=Me:GetLoc(pzone,px,py)
						if not newpos then
							mod.f:Print("ERROR relocating! pos="..(pos or "nil")..", pzone="..(pzone or "nil").." ("..(x or "nil").."/"..(y or "nil")..")\zone was "..(zone or "nil")..", nr="..(nr or "nil"))
						else
							db.targets.list[newpos]={}
							for who,when in pairs(whowhen) do 
								db.targets.list[newpos][who]=when 
							end
							db.targets.list[pos]=nil
							changed=changed+1
						end
					end
				end
			end
			mod.f:PrintR(L.msg.zoneidchanged,{zone=zone,old=nr,new=L.locator.zones[zone],changes=changed})
			db.locator.unknown[zone]=nil
		else
			L.locator.zones[zone]=nr
			C.zones[nr]={}
			mod.f:PrintR(L.msg.unknownzone,{zone=zone,nr=nr})
		end
	end

	
	for zone,nr in pairs(L.locator.zones) do 
		if nr>0 and not zone_name[nr] then zone_name[nr]=zone end
	end
		
	widget.postext=mod.ui:addText(myname.."_Pos_Text",65,10,fwinfo.frame,"TOPRIGHT","TOPRIGHT",-4,-16,11,"RIGHT")
	widget.speed=mod.ui:addText(myname.."_Pos_Text",32,10,fwinfo.frame,"BOTTOMRIGHT","TOPRIGHT",-4,2,11,"RIGHT")

	mod.core:Register4Event("TICK","loc",Me.Tick)
end

-- --------------------------------------------------------------------------------------
-- TICK
-- --------------------------------------------------------------------------------------
function Me:Tick(elapsed)
	local x,y,zone,pos,zone_nr=Me:GetLoc()
	if not zone then return end
	widget.postext:SetText(format("%.1f",x).."/"..format("%.1f",y))
	local speed=33.333*Me:Distance(pos,tick.last or pos)/elapsed
	tick.ix=tick.ix+1
	if tick.ix>5 then 
		tick.ix=1 
		tick.tack=1-tick.tack
	end
	tick.sum=tick.sum+speed-(tick.list[tick.ix] or 0)
	tick.list[tick.ix]=speed
	tick.last=pos
	widget.speed:SetText(format("%.0f",math.max(0,100*tick.sum/5)))

	if zone and zone~=tick.zone then
		tick.zone=zone
		for modname,info in pairs(interested.zone) do 
			if info.run then interested.zone[modname].list=info.run(does_not_go_to_method,zone,zone_nr) end
		end
	end
	
	-- interested.zone[name]={list={[1]={pos=,text=,mindist=,maxdist=,icon=},[2]={}...}}
	local mmradius=Minimap:GetWidth() / 2
	local mapscalex,mapscaley,mmscalex,mmscaley=Me:GetScale(zone,Minimap:GetZoom())
	local icon=0
	for modname,info in pairs(interested.zone) do
		if info.list then
			for nr,data in pairs(info.list) do
				local distance=Me:Distance(pos,data.pos)
				if distance>=(data.mindist or MMMINDIST) and distance<=(data.maxdist or MMMAXDIST) then
					local mmpoint={x=0,y=mmradius-8,color={red=0,green=0,blue=1,opacity=1}}
					local ix,iy,izone,izone_nr=Me:ParseLoc(data.pos)
					if zone_nr==izone_nr then
						local mmx=(ix-x)*mmradius/mmscalex
						local mmy=(y-iy)*mmradius/mmscaley
						local mmdistance=sqrt(mmx * mmx + mmy * mmy)
						if mmdistance < mmradius-8 then
							mmpoint.x=mmx
							mmpoint.y=mmy
						else
							mmpoint.x=mmx*(mmradius-8)/mmdistance
							mmpoint.y=mmy*(mmradius-8)/mmdistance
						end	
						if 		distance<1 then mmpoint.color={red=0,green=1,blue=0,opacity=1}
						elseif 	distance<2 then mmpoint.color={red=1,green=1,blue=0,opacity=1}
						elseif 	distance<3 then mmpoint.color={red=1,green=0,blue=0,opacity=1}
						end
						local endfade=(data.mindist or MMMINDIST)
						local beginfade=(data.mindist or MMMINDIST)+(data.faderange or MMFADERANGE)
						if beginfade>endfade and distance<beginfade then mmpoint.color.opacity=(distance-endfade)/(beginfade-endfade) end
					end
					if tick.tack==1 then mmpoint.color={red=1,green=1,blue=1,transparence=mmpoint.color.opacity} end
					icon=icon+1
					if not widget.mm[icon] then
						widget.mm[icon]=mod.ui:newButton(myname.."_MM_"..icon,12,12,"Minimap",nil,nil,nil,nil,{inherit="Zifnabs_MM"})
						widget.mm.high=icon
					end
					widget.mm[icon]:SetPoint("CENTER","Minimap","CENTER",mmpoint.x,mmpoint.y)
					if data.icon then 
						local iconname="Interface/AddOns/Zifnabs/Icons/"..data.icon..".tga"
						widget.mm[icon]:SetBackdrop(nil)
						widget.mm[icon]:SetNormalTexture(iconname)
						widget.mm[icon]:SetSize(12,12)
					else
						widget.mm[icon]:SetBackdrop({bgFile = "Interface/Tooltips/UI-Tooltip-Background", edgeFile = nil});
						widget.mm[icon]:SetNormalTexture(nil)
						widget.mm[icon]:SetSize(8,8)
						widget.mm[icon]:SetBackdropColor(mmpoint.color.red,mmpoint.color.green,mmpoint.color.blue,1)
						widget.mm[icon]:SetAlpha(mmpoint.color.opacity or 1)
					end
					widget.mm[icon].tooltipText=data.text
					widget.mm[icon].click=data.click
					widget.mm[icon].id=data.id
					widget.mm[icon].distance2=data.pos
					widget.mm[icon]:Show()
				end
			end
		end
	end
	for hide=icon+1,(widget.mm.high or 0),1 do widget.mm[hide]:Hide() end
end

-- --------------------------------------------------------------------------------------
-- [DE]REGISTER FOR ZONE/MAP CHANGE
-- --------------------------------------------------------------------------------------
function Me:Register4Zone(name,method) interested.zone[name]={run=method} end
function Me:Register4Map(name,method) interested.map[name]={run=method} end
function Me:DeRegister4Zone(name,method) interested.zone[name]=nil end
function Me:DeRegister4Map(name,method) interested.map[name]=nil end

function Me:WORLDMAP_ZONE_CHANGED(zone_nr)
	local zone=zone_name[zone_nr]
	for modname,info in pairs(interested.map) do 
		if info.run then interested.map[modname].list=info.run(does_not_go_to_method,zone,zone_nr) end
	end
	return interested.map
end

-- --------------------------------------------------------------------------------------
-- SET ZONE/MAP LIST
-- --------------------------------------------------------------------------------------
function Me:SetZoneList(name,list) 
	if not interested.zone[name] then interested.zone[name]={} end
	interested.zone[name].list=list 
end
function Me:SetMapList(name,list) 
	if not interested.map[name] then interested.map[name]={} end
	interested.map[name].list=list 
end

-- --------------------------------------------------------------------------------------
-- GETLOC
-- --------------------------------------------------------------------------------------
function Me:GetLoc(zone, x, y)
	if not zone then 
		zone=GetRealZoneText()
		local umid=C_Map.GetBestMapForUnit("player")
		if umid then x,y=C_Map.GetPlayerMapPosition(umid,"player"):GetXY() 
		else
			x=0
			y=0
		end
	end
	if not zone then return nil end
	local zone_nr=Me:GetZoneNumber(zone)
	x=math.min(POS_MULTI-1,math.floor(POS_MULTI*x+0.5))
	y=math.min(POS_MULTI-1,math.floor(POS_MULTI*y+0.5))
	return x/FRACT,y/FRACT,zone,ZONE_MULTI*zone_nr+POS_MULTI*x+y,zone_nr
end

-- -----------------------------------------------------------------
-- ZONE NAME
-- -----------------------------------------------------------------
function Me:ZoneName(zone_nr)
	return zone_name[zone_nr]
end

-- -----------------------------------------------------------------
-- PARSE LOCATION NUMBER
-- -----------------------------------------------------------------
function Me:ParseLoc(loc)
	local zone_nr=math.floor(loc/ZONE_MULTI)
	loc=loc-zone_nr*ZONE_MULTI
	local x=math.floor(loc/POS_MULTI)
	local y=loc-x*POS_MULTI
	return x/FRACT,y/FRACT,zone_name[zone_nr],zone_nr
end

-- --------------------------------------------------------------------------------------
-- DISTANCE
-- --------------------------------------------------------------------------------------
function Me:Distance(pos1,pos2)
	if not pos2 then dummyx,dummyy,dummyz,pos2,dummynr=Me:GetLoc() end
	local x1,y1,zone1,zonenr1=Me:ParseLoc(pos1)
	local x2,y2,zone2,zonenr2=Me:ParseLoc(pos2)
	local distance=99999
	if zone1==zone2 then
		local scalex,scaley=Me:GetScale(zone1)
		local vx=(x1-x2)/scalex
		local vy=(y1-y2)/scaley
		distance=math.sqrt(vx*vx+vy*vy)
	end
	return distance
end

-- --------------------------------------------------------------------------------------
-- GET SCALE
-- --------------------------------------------------------------------------------------
function Me:GetScale(zone)
	if not zone then zone=GetRealZoneText() end
	zone=Me:GetZoneNumber(zone)
	local scalex=DEFAULTSCALE.X
	local scaley=DEFAULTSCALE.Y
	if C.zones[zone] then 
		scalex=C.zones[zone].x or DEFAULTSCALE.X
		scaley=C.zones[zone].y or DEFAULTSCALE.Y
	end
	if zoomlevel then
		-- TODO zoomlevel berücksichtigen
	end
	local zoom=Minimap:GetZoom()
	return scalex,scaley,scalex/ZOOMSCALE[zoom],scaley/ZOOMSCALE[zoom]
end

function Me:GetZoneNumber(zone)
	local nr=nil
	if L.locator.zones[zone] then nr=math.abs(L.locator.zones[zone])
	elseif db.locator.unknown[zone] then nr=db.locator.unknown[zone] end
	if not nr then
		nr=db.locator.nextunknownzone or 10000
		db.locator.nextunknownzone=nr+1
		db.locator.unknown[zone]=nr
		L.locator.zones[zone]=nr
		C.zones[nr]={}
		mod.f:PrintR(L.msg.unknownzone,{zone=zone,nr=nr},"buzzer.mp3")
	end
	return nr
end